@@ -22,18 +22,6 @@ let display_errors loc f p =
   with
     Underivable msg | Failure msg -> fatal_error loc msg
 
-(** ... *)
-
-let contains_tvars, contains_tvars_decl =
-  let o = object
-     inherit [bool] fold as default
-     method crush = List.exists F.id
-     method expr = function
-       | `Param _ -> true
-       | e -> default#expr e
-  end in (o#expr, o#decl)
-
-
 (** *)
 
 let instantiate, instantiate_repr =
@@ -41,6 +29,8 @@ let instantiate, instantiate_repr =
     inherit transform as super
     method expr = function
     | `Param (name, _) -> lookup name
+    | `GParam ((name, _), e) ->
+         if not (contains_tvars e) then e else lookup name
     | e                -> super # expr e
   end in
   (fun (lookup : name -> expr) -> (o lookup)#expr),
@@ -119,6 +109,11 @@ module AstHelpers(Loc : Loc) = struct
     | [x] -> x
     | x::xs -> Ast.ExTup (_loc, List.fold_left (fun e t -> Ast.ExCom (_loc, e,t)) x xs)
 
+  let tuple_patt : Ast.patt list -> Ast.patt = function
+    | [] -> <:patt< () >>
+    | [x] -> x
+    | x::xs -> Ast.PaTup (_loc, List.fold_left (fun e t -> Ast.PaCom (_loc, e,t)) x xs)
+
   let tuple ?(param="v") n : string list * Ast.patt * Ast.expr =
     let v n = Printf.sprintf "%s%d" param n in
       match n with
@@ -138,7 +133,20 @@ module AstHelpers(Loc : Loc) = struct
 
   (** *)
 
+  let rec lident qname =
+    match qname with
+      | [] -> invalid_arg "ident"
+      | [t] -> <:ident< $lid:t$ >>
+      | t::ts -> <:ident< $uid:t$.$lident ts$ >>
+
   let cast_pattern argmap ?(param="x") ty =
+    match ty with
+    | `Constr (id, _) ->
+      let ty = Untranslate.expr (instantiate_modargs argmap ty) in
+      (<:patt< #$id:lident id$ as $lid:param$ >>,
+       <:expr< >>,
+       <:expr< $lid:param$ >>)
+    | ty ->
     let ty = Untranslate.expr (instantiate_modargs argmap ty) in
     (<:patt< $lid:param$ >>,
      <:expr<
@@ -183,7 +191,13 @@ module AstHelpers(Loc : Loc) = struct
 
 end
 
-module Generator(Loc: Loc)(Desc : ClassDescription) = struct
+module type InnerClassDescription = sig
+  include ClassDescription
+  val find_predefined: Type.qname -> Type.qname
+  val depends: (module DepClassBuilder) list
+end
+
+module InnerGenerator(Loc: Loc)(Desc : InnerClassDescription) = struct
 
   (** How does it works ?
 
@@ -227,7 +241,9 @@ module Generator(Loc: Loc)(Desc : ClassDescription) = struct
 
   *)
 
-  module Helpers = AstHelpers(Loc)
+  module Loc = Loc
+  module AstHelpers = AstHelpers(Loc)
+  module Helpers = AstHelpers
   module Untranslate = Helpers.Untranslate
 
   open Loc
@@ -250,9 +266,21 @@ module Generator(Loc: Loc)(Desc : ClassDescription) = struct
   let cast_pattern ctxt ?param ty =
     Helpers.cast_pattern ctxt.argmap ?param ty
   let instantiate_modargs_repr ctxt = instantiate_modargs_repr ctxt.argmap
+  let instantiate_gparam e =
+    let map = object (self)
+      inherit Type.transform as super
+      method expr e = match e with
+      | `GParam (p, e) ->
+          if Type.contains_tvars e then
+            `Param p
+          else
+            e
+      | e -> super#expr e
+    end in
+    map#expr e
 
   let import_depend ctxt ty depend =
-    let module Dep = (val depend : FullClassBuilder)(Loc) in
+    let module Dep = (val depend : DepClassBuilder)(Loc) in
     let argmap =
       NameMap.map (fun qname -> qname @ [Dep.classname]) ctxt.argmap
     and mod_insts =
@@ -345,6 +373,7 @@ module Generator(Loc: Loc)(Desc : ClassDescription) = struct
     method expr ctxt (ty: Type.expr) =
       match ty with
       | `Param p    ->                   (self#param      ctxt p)
+      | `GParam p   ->                   (self#gparam     ctxt p)
       | `Object o   -> self#wrap ctxt ty (self#object_    ctxt o)
       | `Class c    -> self#wrap ctxt ty (self#class_     ctxt c)
       | `Label l    -> self#wrap ctxt ty (self#label      ctxt l)
@@ -353,9 +382,11 @@ module Generator(Loc: Loc)(Desc : ClassDescription) = struct
       | `Tuple t    -> self#wrap ctxt ty (self#tuple      ctxt t)
 
     method rhs ctxt subst (tname, params, rhs, constraints, _ : Type.decl) =
+
       let params =
 	List.map (substitute_expr subst) (List.map (fun p -> `Param p) params)
       in
+
       let ty = `Constr([tname], params) in
       let rhs = substitute_rhs subst rhs in
       match rhs with
@@ -364,6 +395,8 @@ module Generator(Loc: Loc)(Desc : ClassDescription) = struct
 				^ " cannot be derived for private types"))
         | `Fresh (eq, Sum summands, _) ->
 	    self#wrap ctxt ty (self#sum ?eq ctxt tname params constraints summands)
+        | `Fresh (eq, GSum (tname', summands), _) ->
+	    self#wrap ctxt ty (self#gsum ?eq ctxt tname params constraints summands)
         | `Fresh (eq, Record fields, _) ->
 	    self#wrap ctxt ty (self#record ?eq ctxt tname params constraints fields)
         | `Expr e -> self#expr ctxt e
@@ -374,6 +407,9 @@ module Generator(Loc: Loc)(Desc : ClassDescription) = struct
     method param ctxt (name, _) =
       <:module_expr< $id:Untranslate.qName (NameMap.find name ctxt.argmap)$ >>
 
+    method gparam ctxt (p, e) =
+      if Type.contains_tvars e then self#param ctxt p else self#expr ctxt e
+
     method constr ctxt (qname, params) =
       match qname with
       | [tname] when EMap.mem (tname,params) ctxt.mod_insts ->
@@ -382,7 +418,7 @@ module Generator(Loc: Loc)(Desc : ClassDescription) = struct
       | _ ->
 	(* External module: apply classical functor. *)
         let qname =
-	  try List.assoc qname Desc.predefs
+	  try Desc.find_predefined qname
 	  with Not_found -> qname in
 	List.fold_left
 	  (fun m p -> <:module_expr< $m$ ($self#expr ctxt p$) >>)
@@ -409,6 +445,8 @@ module Generator(Loc: Loc)(Desc : ClassDescription) = struct
 	    field list -> Ast.str_item list
     method virtual tuple: context -> expr list -> Ast.str_item list
 
+    method gsum ?eq ctxt tname params constraints gsummands =
+      raise (Underivable (Desc.classname ^ " cannot be derived for GADT"))
     method object_   _ o =
       raise (Underivable (Desc.classname ^ " cannot be derived for object types"))
     method class_    _ c =
@@ -468,6 +506,28 @@ module Generator(Loc: Loc)(Desc : ClassDescription) = struct
 
       let cluster_argmap = make_argmap cluster.Clusters.params in
 
+      let rec wrap_local_types (args : expr list) body =
+#if ocaml_version < (4, 00)
+        body
+#else
+        match args with
+        | [] -> body
+        | `Param (arg, _) :: args
+        | `GParam ((arg, _),_) :: args ->
+          let id = "deriving_" ^ random_id 8 ^ "_" ^ arg in
+            let pat =
+              (* (module $M_arg$ : $class_sig ...$ ) *)
+              Ast.PaTyc(_loc, Ast.PaMod (_loc, "M_" ^ arg),
+                        Ast.TyPkg(_loc,
+                                  gen#class_sig NameMap.empty (`Constr ([id], []))))
+            in
+            wrap_local_types args
+              <:expr< (fun (type $lid:id$) -> (function $pat$ -> $body$))
+                     (module $uid:"M_"^arg$) >>
+        | _ -> assert false
+#endif
+      in
+
       let generate_instance (tname, eparams as inst) =
 	let mod_insts =
 	  EMap.mapi
@@ -482,7 +542,8 @@ module Generator(Loc: Loc)(Desc : ClassDescription) = struct
 	  List.find (fun (tn,_,_,_,_) -> tname = tn) decls in
 	let subst = create_subst params eparams in
 	let body = gen#pack ctxt.argmap ty (gen#rhs ctxt subst decl) in
-	<:binding< $lid:EMap.find inst fun_names$ = lazy $body$ >>
+        let id = EMap.find inst fun_names in
+	id, <:expr< lazy (ignore($lid:id$); $body$) >>
       in
 
       let generate_functor (tname,params,_,_,_ as decl) =
@@ -544,17 +605,35 @@ module Generator(Loc: Loc)(Desc : ClassDescription) = struct
 	>>
       in
 
+      let is_gadt = function
+        | (_,_,`Fresh(_,GSum _,_),_,_) -> true
+        | _ -> false in
+      let contains_gadt c = List.exists is_gadt c.Clusters.decls in
+
       if cluster.Clusters.instances <> [] then
-	let intances =
+        let inst_exprs = List.map generate_instance cluster.Clusters.instances in
+        let bindings =
+          List.map (fun (id, e) -> <:binding< $lid:id$ = $e$ >>) inst_exprs in
+        let items =
+          if not (contains_gadt cluster) then
+            <:str_item< let rec $list:bindings$ >>
+          else
+            let e = List.map (fun (id, _) -> <:expr< $lid:id$ >>) inst_exprs in
+            let p = List.map (fun (id, _) -> <:patt< $lid:id$ >>) inst_exprs in
+            let body =
+              wrap_local_types
+                (List.map (fun p -> `Param p) cluster.Clusters.params)
+                <:expr< let rec $list:bindings$ in $Helpers.tuple_expr e$ >> in
+            <:str_item< let $Helpers.tuple_patt p$ = $body$ >>
+        in
+	let mod_expr =
 	  List.fold_right
 	    (add_functor_param cluster_argmap)
 	    cluster.Clusters.params
-	    <:module_expr< struct
-	      let rec $list:List.map generate_instance cluster.Clusters.instances$
-	    end >>
+	    <:module_expr< struct $items$ end >>
 	in
 	<:str_item<
-          module $uid:cluster_name$ = $intances$
+          module $uid:cluster_name$ = $mod_expr$
 	  $list:List.map generate_functor cluster.Clusters.decls$
         >>
       else
@@ -595,37 +674,75 @@ module Generator(Loc: Loc)(Desc : ClassDescription) = struct
 
 end
 
-
-let derive_str _loc decls (desc, class_builder) =
+let derive_str _loc decls class_builder =
   let module Loc = struct let _loc = _loc end in
-  let module Desc = (val desc : ClassDescription) in
-  let module Class = (val class_builder : ClassBuilder)(Loc) in
+  let module Class = (val class_builder : InnerClassBuilder)(Loc) in
   display_errors _loc Class.generate decls
 
-let derive_sig _loc decls (desc, class_builder) =
+let derive_sig _loc decls class_builder =
   let module Loc = struct let _loc = _loc end in
-  let module Desc = (val desc : ClassDescription) in
-  let module Class = (val class_builder : ClassBuilder)(Loc) in
+  let module Class = (val class_builder : InnerClassBuilder)(Loc) in
   display_errors _loc Class.generate_sigs decls
 
-
-let generators = Hashtbl.create 15
-let hashtbl_add (desc, _ as deriver) =
+let generators : (string, (module InnerClassBuilder)) Hashtbl.t =
+  Hashtbl.create 15
+let hashtbl_add desc deriver =
   let module Desc = (val desc : ClassDescription) in
   Hashtbl.add generators Desc.classname deriver
 let register_hook = ref [hashtbl_add]
 let add_register_hook f = register_hook := f :: !register_hook
-let register (desc, deriver) =
-  List.iter (fun f -> f (desc, deriver)) !register_hook
+let register desc deriver =
+  List.iter (fun f -> f desc deriver) !register_hook
 let find classname =
   try Hashtbl.find generators classname
   with Not_found -> raise (NoSuchClass classname)
 let is_registered classname = Hashtbl.mem generators classname
 
-module Register
-    (Desc : ClassDescription)
+module MakeInnerDesc(Desc : Defs.ClassDescription) =
+struct
+  include Desc
+  let predefs_tbl : (Type.qname, Type.qname) Hashtbl.t = Hashtbl.create 17
+  let find_predefined qname = Hashtbl.find predefs_tbl qname
+  let register_predefined ty instance = Hashtbl.add predefs_tbl ty instance
+  let () = List.iter (fun (a, b) -> register_predefined a b) predefs
+end
+
+module RegisterClass
+    (Desc : Defs.ClassDescription)
     (MakeClass : ClassBuilder) = struct
 
- let _ = register ((module Desc : ClassDescription), (module MakeClass : ClassBuilder))
+  module InnerDesc = MakeInnerDesc(Desc)
+  module Builder(Loc: Loc) = MakeClass(InnerGenerator(Loc)(InnerDesc))
+
+  let _ = register (module Desc : Defs.ClassDescription) (module Builder : InnerClassBuilder)
+
+  let register_predefs = InnerDesc.register_predefined
 
 end
+
+module RegisterFullClass
+    (Desc : Defs.ClassDescription)
+    (MakeClass : FullBuilder) = struct
+
+  module InnerDesc = MakeInnerDesc(Desc)
+  module Builder(Loc: Loc) = MakeClass(InnerGenerator(Loc)(InnerDesc))
+
+  let _ = register (module Desc : Defs.ClassDescription) (module Builder : InnerClassBuilder)
+
+  let depends = (module Builder : DepClassBuilder)
+  let register_predefs = InnerDesc.register_predefined
+
+end
+
+(* Compat with <= 0.4-ocsigen *)
+
+module Register
+    (Desc : Defs.ClassDescription)
+    (MakeClass : InnerClassBuilder) = struct
+
+  let _ = register (module Desc : Defs.ClassDescription) (module MakeClass : InnerClassBuilder)
+
+end
+
+module Generator(Loc : Loc)(Desc : ClassDescription) =
+  InnerGenerator(Loc)(MakeInnerDesc(Desc))
